//
// Helper functions
//

function computeCumulativeOffset(element) {
	let top = 0, left = 0;
	do {
		top += element.offsetTop || 0;
		left += element.offsetLeft || 0;
		element = element.offsetParent;
	} while (element);

	return {
		top: top,
		left: left
	};
}

function htmlDecode(str) {
	let map = {amp: '&', lt: '<', gt: '>', quot: '"', '#039': "'", '#x27': "'"};
	return str.replace(/&([^;]+);/g, (m, c) => map[c]);
}

//
// UNDO element classes
//

class UndoElement {
	constructor(type) {
		this.type = type;
	}
}

class UndoPath extends UndoElement {
	constructor(eraser, penSize, coords) {
		super('path');
		this.eraser = eraser;
		this.penSize = penSize;
		this.coords = coords;
	}
}

class UndoTextarea extends UndoElement {
	constructor(type, id) {
		super(type);
		this.id = id;
	}
}

class UndoTextareaCreate extends UndoTextarea {
	constructor(id, x, y, string) {
		super('create', id);
		this.x = x;
		this.y = y;
		this.string = string;
	}
}

class UndoTextareaDelete extends UndoTextarea {
	constructor(id) {
		super('delete', id);
	}
}

class UndoTextareaChange extends UndoTextarea {
	constructor(id, string) {
		super('change', id);
		this.string = string;
	}
}

class UndoTextareaMove extends UndoTextarea {
	constructor(id, newX, newY) {
		super('move', id);
		this.newX = newX;
		this.newY = newY;
	}
}

//
// Textarea class
//

class TextAreaObject {
	constructor(div, textarea, id, page, problem, subproblem) {
		this.id = id;
		this.textarea = textarea;
		this.div = div;
		this.page = page;
		this.problem = problem;
		this.subproblem = subproblem;
	}

	resizeTextarea() {
		this.textarea.style.height = 'auto';
		this.textarea.style.width = '100%';

		// +2 because otherwise the text was shifted a little on the ipad
		this.textarea.style.height = (this.textarea.scrollHeight + 2) + 'px';
		this.textarea.style.width = (this.textarea.scrollWidth + 2) + 'px';
	}
}


//
// Canvas, Context and remaining global variables
//

let exam_data = {
	maxPage: null,
	pages: [{
		landscape: true,
		layers: []
	}],
};

let isReadonly = false;
let lockToken = '';
let lockRetries = 0;
const lockMaxRetries = 5;
const keepAliveInterval = 50; // In Seconds

let imagesLoaded = [];
let initialLayer = [];
initialLayer[1] = [];
initialLayer[2] = [];
initialLayer[3] = [];
let loadedLayers = 2;
let slidesLoaded = [];
let slidesLoading = [];
let rawImageLoaded = [];
let rawImageLoading = [];

let imgSize = [];
// [ {width: ..., height: ...}, ...]

const defaultFontSize = 45;
let textAreaFontSize = 0;
let canvasFontSize = 0;

let container = null;
const allSlides = [];

const mainCanvas = [];
const mainContext = [];

let maxSlide = 0;

// Offset of the canvas container used when drawing
let pos = null;
let drawing = false;
let scaleFactorContainerToCanvas = [];
const inactiveOpacity = '0.3';
const activeOpacity = '1';

const canvasOpacity = '0.8';
const canvasOpacityLow = '0.2';

//
// Variables for changing size, color, ...
//

let currentProblem = null;
let currentSubproblem = null;
let activeProblems = [];
let problems = [];
let boxidToProblem = [];
let unsavedChanges = false;

let currentSlide = 0;
let dragStart = 0;
let dragID = 0;
let dragTime = 0;

let lastPoint = null;
let lastCenterPoint = null;
let currentPath = null;
let drawID = 0;

let undoArray = [];
let redoArray = [];
let tempTextAreaContent = '';

let textAreaArray = [];
let textAreaCounter = [];
let textAreaCurrentCounter = [];

let eraserSizeFactor = 20;
let currentPenSize = 2;

let eraserActive = false;
let textMode = false;

let fullscreen = false;
let solutionOpacity = 1;
let showIndividualSolution = false;

let currentScale = 1;
let zoomStart = [];
let translateStart = [];
let currentTranslate = [];
let translateX = 0;
let translateY = 0;
let isZooming = false;
let isTranslating = false;
let doubleTabStart = 0;
let doubleTabCoords = [0, 0];

let clickedOnTextarea = false;
let currentCutTextarea = -1;
let cutTextareaUndoArray = null;
let currentUser = 1;
let correctionPass = 1;

const colorRed = TUMexamColors.Red;
const colorGreen = TUMexamColors.Green;
const colorPink = TUMexamColors.Purple;
const colorOrange = TUMexamColors.Yellow;

let colors = ['black', 'blue', colorRed, colorGreen, colorPink];

let fingerDraw = false;
let saveIsRunning = false;

// Icons
const iconSpinner = '<i class="fas fa-spinner fa-pulse fa-5x fa-fw"></i>';
const iconCut = '<i class="fas fa-cut fa-fw" aria-hidden="true"></i>';
const iconAccept = '<i class="fas fa-check fa-fw" aria-hidden="true"></i>';
const iconClose = '<i class="fas fa-times fa-fw" aria-hidden="true"></i>';
const iconPaste = '<i class="fas fa-clipboard fa-fw" aria-hidden="true"></i>';
const iconSave = '<i class="fas fa-save" aria-hidden="true"></i>';
const iconSpinnerSmall = '<i class="fas fa-spinner fa-pulse fa-fw"></i>';
const iconLockClosed = '<i class="fas fa-lock"></i>';
const iconLockOpen = '<i class="fas fa-unlock"></i>';

//
// Initialize
//

export function initialize(init_data) {

	$('#toolbarLog').on('click', evt => {
		$('#toolbarLogModal').modal("show");
	})

	getCorrectionPreferences();
	isReadonly = init_data.readonly;
	lockToken = init_data.lockToken;

	let url = '/data'

	let options = {
		method: 'GET',
		mode: 'same-origin',
		cache: 'no-cache',
		credentials: 'include',
	};

	fetch(url, options).then(function (res) {
		if (!res.ok)
			throw res;
		return res.json();
	}).then(data => {
		exam_data = data;
		setup(init_data);
	}).catch(res => {
		console.log(res);
		let container = document.getElementById('sliderContainer');
		let p = document.createElement('p');
		if(res.status !== undefined || res.statusText !== undefined)
			p.innerHTML = `Could not contact server<br>${res.status} ${res.statusText}`;
		else
			p.innerHTML = `An unknown error occurred. Please reload the page.`;
		container.appendChild(p);

		$('#toolbar .option').addClass('disabled');
	});

}

function setup(init_data) {

	// max page number
	maxSlide = exam_data.maxPage;

	correctionPass = init_data.correction_pass;
	currentProblem = init_data.problem;
	currentSubproblem = init_data.subproblem;
	currentUser = correctionPass;
	currentSlide = init_data.initialPage;

	container = document.getElementById('sliderContainer');

	if (isReadonly) {
		readonlyToolbar();
		updateLog('Readonly', 'Exam set to readonly.', 'alert-warning');
	} else {
		for (let i = 0; i < currentSubproblem.length; i++) {
			activeProblems.push(currentProblem + currentSubproblem[i]);
		}
	}

	undoArray[correctionPass] = [];
	redoArray[correctionPass] = [];

	let eraserElement = document.createElement('div');
	eraserElement.id = 'eraser';
	container.appendChild(eraserElement);

	createPagination();
	setPagination();

	setupCanvas();
	createSlides();

	addEventListeners();

	loadSlide(currentSlide);
	loadSlide(currentSlide + 1);
	loadSlide(currentSlide - 1);

	toggleFullscreen();

	if(!isReadonly) {
		lockKeepAlive();
	}
}


//
// Scale canvas according to image size
//

function attachCanvas(page) {

	if (!slidesLoaded[page])
		return;

	mainCanvas[correctionPass].width = imgSize[page].width;
	mainCanvas[correctionPass].height = imgSize[page].height;

	let imageScale = Math.max(imgSize[page].width, imgSize[page].height) / 3500;
	canvasFontSize = defaultFontSize * imageScale;
	textAreaFontSize = defaultFontSize * scaleFactorContainerToCanvas[page] * imageScale;

	setupContext(mainCanvas[correctionPass].getContext('2d'), correctionPass);
	replayUndoArraySlide(mainCanvas[correctionPass], page, correctionPass);
	allSlides[page].appendChild(mainCanvas[correctionPass]);

	setEraserContext(eraserActive);
	setPenSize(currentPenSize);
}

function setupContext(ctx, user) {
	ctx.lineCap = 'round';
	ctx.lineJoin = 'round';
	ctx.font = canvasFontSize + 'px sans-serif';
	ctx.fillStyle = colors[user];
	ctx.strokeStyle = colors[user];
}


function setupCanvas() {

	mainCanvas[correctionPass] = document.createElement('canvas');
	mainCanvas[correctionPass].classList = 'canvasDraw';
	mainContext[correctionPass] = mainCanvas[correctionPass].getContext('2d');
}

//
// Setup Slides and Images
//

function createSlides() {

	for (let i = 1; i <= maxSlide; i++) {

		let slide = document.createElement('div');
		slide.classList.add('singleSlide');
		slide.classList.add('noDisplay');

		// Create spinner while loading the exam page
		let spinner = document.createElement('div');
		spinner.classList.add('load_image_spinner');
		spinner.innerHTML = iconSpinner;
		slide.appendChild(spinner);

		if (i > currentSlide)
			slide.style.transform = 'translate(100%,0)';
		else if (i < currentSlide)
			slide.style.transform = 'translate(-100%,0)';

		container.appendChild(slide);

		allSlides[i] = slide;

		initialLayer[correctionPass][i] = new Image();

		imagesLoaded[i] = 0;
		slidesLoaded[i] = false;
		slidesLoading[i] = false;
		rawImageLoaded[i] = false;
		rawImageLoading[i] = false;

		undoArray[correctionPass][i] = [];
		redoArray[correctionPass][i] = [];
	}

	textAreaArray[correctionPass] = [];
	textAreaCounter[correctionPass] = [];
}

function loadSlide(pageNumber) {

	if (slidesLoading[pageNumber] || pageNumber < 1 || pageNumber > maxSlide)
		return;

	slidesLoading[pageNumber] = true;

	let options = {
		method: 'GET',
		mode: 'same-origin',
		credentials: 'include',
		headers: {
			'Content-Type': 'image/jpeg',
		},
	};

	let url = `/image/${pageNumber}`;

	fetch(url, options).then(function (response) {
		if (response.ok)
			response.blob().then(function (blob) {
				let img = new Image();
				img.onload = function () {
					allSlides[pageNumber].appendChild(img);

					imgSize[pageNumber] = {
						width: img.naturalWidth,
						height: img.naturalHeight
					};

					if (++imagesLoaded[pageNumber] === loadedLayers)
						setupSlide(pageNumber);
				};
				img.src = URL.createObjectURL(blob);
			});
		else if (response.status === 404) {
			response.blob().then(function (blob) {
				let img = document.createElement('img');
				img.onload = function () {
					removeSpinner(pageNumber);
					setContainerHeight(pageNumber);
					allSlides[pageNumber].classList.remove('noDisplay');
				};
				img.src = URL.createObjectURL(blob);
				allSlides[pageNumber].appendChild(img);
				allSlides[pageNumber].classList.add('scanNotFound');
			});
		} else {
			throw response;
		}
	}).catch(function (response) {
		removeSpinner(pageNumber);
		let p = document.createElement('p');
		p.innerHTML = `Could not load page<br>${response.status}: ${response.statusText}`;
		allSlides[pageNumber].appendChild(p);
	});

	initialLayer[correctionPass][pageNumber] = new Image();
	if (++imagesLoaded[pageNumber] === loadedLayers)
		setupSlide(pageNumber);
}

function setupSlide(pageNumber) {
	allSlides[pageNumber].classList.remove('noDisplay');
	slidesLoaded[pageNumber] = true;
	scaleFactorContainerToCanvas[pageNumber] = container.clientWidth / imgSize[pageNumber].width;
	setToolbarIcons();
	setContainerHeight(pageNumber);
	removeSpinner(pageNumber);
	if (pageNumber === currentSlide) {
		attachCanvas(pageNumber);
	}
}

function removeSpinner(pageNumber) {
	let loadImageSpinner = allSlides[pageNumber].getElementsByClassName('load_image_spinner');
	if (loadImageSpinner[0])
		loadImageSpinner[0].parentNode.removeChild(loadImageSpinner[0]);
}

function setContainerHeight(page) {
	if (page !== currentSlide)
		return;
	container.style.height = allSlides[page].offsetHeight + 'px';
}

//
// Generate Textboxes
//

function setTextareaMaxSize(textarea, posX, posY) {
	//textarea.style.maxWidth = Math.max(150, container.clientWidth * (1 - posX / 100)) + 'px';
	//let headerHeight = textarea.parentNode.getElementsByClassName('textAreaHeader')[0].clientHeight;
	//textarea.style.maxHeight = Math.max(60, container.clientHeight * (1 - posY / 100) - headerHeight) + 'px';
}

function generateTextarea(problem, subproblem, x, y, string = '') {

	// Decode escaped string from database
	string = htmlDecode(string);

	// switch to writing mode
	if (!textMode) {
		toggleTextMode();
	}

	let div = document.createElement('div');
	div.classList.add('textArea');
	div.style.left = x + '%';
	div.style.top = y + '%';
	allSlides[currentSlide].appendChild(div);

	let header = document.createElement('div');
	div.appendChild(header);
	header.classList.add('textAreaHeader');

	let textarea = document.createElement('textarea');
	textarea.classList.add('form-control')
	textarea.setAttribute('wrap', 'off');
	textarea.style.fontSize = textAreaFontSize + 'px';
	textarea.style.color = colors[currentUser];
	textarea.value = string;
	div.appendChild(textarea);

	let name = document.createElement('select');
	header.appendChild(name);
	name.classList.add('textAreaOption', 'form-control', 'form-select');
	$.each(activeProblems, (i, v) => {
		$(`<option value="${v}" ${ (problem + subproblem === v) && 'selected' }>${v}</option>`).appendTo(name);
	});

	function createHeaderButton(icon, color) {
		let elem = document.createElement('button');
		header.appendChild(elem);
		elem.classList.add('textAreaOption', 'btn');
		elem.innerHTML = icon;
		elem.style.color = color;
		return elem;
	}

	let cut = createHeaderButton(iconCut, colorOrange);
	let accept = createHeaderButton(iconAccept, colorGreen);
	let close = createHeaderButton(iconClose, colorRed);
	let paste = createHeaderButton(iconPaste, colorOrange);

	let textAreaID = textAreaArray[currentUser].length;
	let textAreaObj = new TextAreaObject(div, textarea, textAreaID, currentSlide, problem, subproblem);
	textAreaArray[currentUser].push(textAreaObj);
	textAreaCounter[currentUser].push(0);
	textAreaObj.resizeTextarea();

	setTextareaMaxSize(textarea, x, y);
	addUndoElement(new UndoTextareaCreate(textAreaID, x, y, string));
	textarea.focus();

	$(name).on('change', evt => {
		let result = $(name).val().match('(?<problem>[0-9]+)(?<subproblem>[a-z]+)');
		textAreaObj.problem = result.groups.problem;
		textAreaObj.subproblem = result.groups.subproblem;
	});

	function addMoveObject(x, y) {
		let newX = 100 * x;
		let newY = 100 * y;

		addUndoElement(new UndoTextareaMove(textAreaID, newX, newY));
		textAreaArray[currentUser][textAreaID].resizeTextarea();
		setTextareaMaxSize(textarea, newX, newY);
	}

	textarea.addEventListener('focus', function () {
		tempTextAreaContent = textarea.value;
	});
	textarea.addEventListener('input', function () {
		textAreaArray[currentUser][textAreaID].resizeTextarea();
	});
	textarea.addEventListener('blur', function () {

		// Skip if nothing was changed
		if (tempTextAreaContent === textarea.value)
			return;

		addUndoElement(new UndoTextareaChange(textAreaID, textarea.value));
		//textarea.style.height = 'initial';
		textAreaArray[currentUser][textAreaID].resizeTextarea();
	});

	$(close).bind('click touchstart', (evt) => {
		div.style.top = '2000%';
		textarea.value = '';
		addUndoElement(new UndoTextareaDelete(textAreaID));
		evt.stopPropagation();
	});

	$(accept).bind('click touchstart', (evt) => {
		toggleTextMode();
		evt.stopPropagation();
	});

	$(cut).bind('click touchstart', (evt) => {
		// In case another textarea is already cut
		insertTextarea();
		cutTextarea(textAreaID, currentUser, currentSlide);
		evt.stopPropagation();
	});

	$(paste).bind('click touchstart', (evt) => {
		insertTextarea();
		evt.stopPropagation();
	});

	// Make textarea draggable
	$(div).draggable({
		cancel: ".textAreaOption",
		containment: container,
		stop: function(evt) {
			let $this = $(this);
			let $parent = $this.parent();
            let thisPos = $this.position();
            let parentPos = $parent.position();

            let x = thisPos.left - parentPos.left;
            let y = thisPos.top - parentPos.top;
            addMoveObject(x / $parent.width(), y /$parent.height());
		},
	});

	// Automatically set focus on textarea when clicked
	$(textarea).on('click', evt => {
		textarea.focus();
	});

	// Handle touch textarea drags
	function touchHandler(event) {
		let touch = event.changedTouches[0];

		let simulatedEvent = document.createEvent("MouseEvent");
			simulatedEvent.initMouseEvent({
			touchstart: "mousedown",
			touchmove: "mousemove",
			touchend: "mouseup"
		}[event.type], true, true, window, 1,
			touch.screenX, touch.screenY,
			touch.clientX, touch.clientY, false,
			false, false, false, 0, null);

		touch.target.dispatchEvent(simulatedEvent);
		event.stopPropagation();
	}

	$(div).on("touchstart", touchHandler);
    $(div).on("touchmove", touchHandler);
    $(div).on("touchend", touchHandler);
    $(div).on("touchcancel", touchHandler);
}

function cutTextarea(id, user, page) {

	currentCutTextarea = id;

	let commands = [];
	let newUndoArray = [];
	let newRedoArray = [];

	for (let i = 0; i < undoArray[user][page].length; i++) {
		let obj = undoArray[user][page][i];

		if (obj.type !== 'path' && obj.id === id) {
			commands.push(obj);
		} else {
			newUndoArray.push(obj);
		}
	}

	for (let i = 0; i < redoArray[user][page].length; i++) {
		let obj = redoArray[user][page][i];

		if (obj.type !== 'path' && obj.id === id) {

		} else {
			newRedoArray.push(obj);
		}
	}

	let textAreaDiv = textAreaArray[user][id].div;
	textAreaDiv.classList.add('cut');
	container.append(textAreaDiv);
	undoArray[user][page] = newUndoArray;
	redoArray[user][page] = newRedoArray;
	cutTextareaUndoArray = commands;
}

function insertTextarea() {
	if (currentCutTextarea < 0)
		return;

	//TODO: paste textarea
	let textAreaDiv = textAreaArray[currentUser][currentCutTextarea].div;
	textAreaDiv.classList.remove('cut');
	allSlides[currentSlide].appendChild(textAreaDiv);

	let textArea = textAreaArray[currentUser][currentCutTextarea].textarea;
	textArea.style.fontSize = textAreaFontSize + 'px';
	textArea.style.color = colors[currentUser];
	textAreaArray[currentUser][currentCutTextarea].resizeTextarea();

	undoArray[currentUser][currentSlide] = undoArray[currentUser][currentSlide].concat(cutTextareaUndoArray);
	currentCutTextarea = -1;
	setToolbarIcons();
}


//
// Draw Functions
//

function startDraw(x, y) {

	if (textMode)
		toggleTextMode();

	startNewPath();
	mainContext[currentUser].beginPath();
	mainContext[currentUser].moveTo(x, y);

	lastPoint = [x, y];     // save points for next step
	lastCenterPoint = [x, y];
	currentPath = new Path(currentPenSize, eraserActive);
	addPointToPath(currentPath, x, y);

	disableGuiOnDraw();
}

function draw(x, y) {

	let cx = (lastPoint[0] + x) / 2;
	let cy = (lastPoint[1] + y) / 2;

	mainContext[currentUser].beginPath();
	mainContext[currentUser].moveTo(lastCenterPoint[0], lastCenterPoint[1]);
	mainContext[currentUser].quadraticCurveTo(lastPoint[0], lastPoint[1], cx, cy);
	mainContext[currentUser].stroke();

	lastPoint = [x, y];
	lastCenterPoint = [cx, cy];
	addPointToPath(currentPath, x, y);
}

function stopDraw(x, y) {

	if (currentPath.coords.length < 2) {
		// ignore the path if it has length 1
		currentPath = null;
		if (clickedOnTextarea)
			toggleTextMode();

	} else {
		// draw normal curve to last point
		mainContext[currentUser].quadraticCurveTo(lastPoint[0], lastPoint[1], x, y);
		mainContext[currentUser].stroke();

		addPointToPath(currentPath, x, y);
		stopPath();
	}

	lastPoint = null;
	lastCenterPoint = null;

	enableGuiOnDraw();
}

function checkIfContains(obj, x, y) {

	// Handle if display is removed
	let display = obj.style.display;
	obj.style.display = 'inherit';
	let rect = obj.getBoundingClientRect();
	obj.style.display = display;
	return x > rect.left && x < rect.right && y > rect.top && y < rect.bottom;
}

//
// Eraser
//

function initializeEraser(x, y) {
	let eraser = document.getElementById('eraser');
	eraser.style.display = 'initial';
	eraser.style.width = currentScale * eraserSizeFactor * 2 * currentPenSize * scaleFactorContainerToCanvas[currentSlide] + 'px';
	eraser.style.height = currentScale * eraserSizeFactor * 2 * currentPenSize * scaleFactorContainerToCanvas[currentSlide] + 'px';

	x -= 0.5 * eraser.offsetWidth;
	y -= 0.5 * eraser.offsetHeight;
	eraser.style.transform = `translate(${x}px, ${y}px)`;
}

function moveEraser(x, y) {
	let eraser = document.getElementById('eraser');

	x -= 0.5 * eraser.offsetWidth;
	y -= 0.5 * eraser.offsetHeight;
	eraser.style.transform = `translate(${x}px, ${y}px)`;
}

function stopEraser() {
	let eraser = document.getElementById('eraser');
	eraser.style.display = 'none';
}

function getDrawCoordinates(evt) {

	let coords = getRelativeCoordinates(evt.pageX, evt.pageY);

	let eventX = coords.x;
	let eventY = coords.y;

	let x = -translateX / 100 + eventX / container.clientWidth / currentScale;
	let y = -translateY / 100 + eventY / container.clientHeight / currentScale;

	x *= imgSize[currentSlide].width;
	y *= imgSize[currentSlide].height;

	return {
		drawX: x,
		drawY: y,
		eraserX: eventX,
		eraserY: eventY
	};

}

function getRelativeCoordinates(eventX, eventY) {

	/*
	let scrollLeft = document.getElementsByClassName('main')[0].scrollLeft;
	let scrollTop = document.getElementsByClassName('main')[0].scrollTop;

	return {
		x: eventX - pos.left + scrollLeft,
		y: eventY - pos.top + scrollTop
	};

	 */
	// TODO: can this function be removed?
	return {
		x: eventX - pos.left,
		y: eventY - pos.top
	};
}

// Mouse Input
function mouseStart(evt) {

	setContainerHeight(currentSlide);
	pos = computeCumulativeOffset(container);

	let drawCoords = getDrawCoordinates(evt);

	drawing = true;
	startDraw(drawCoords.drawX, drawCoords.drawY);
	if (eraserActive)
		initializeEraser(drawCoords.eraserX, drawCoords.eraserY);
}

function mouseMove(evt) {

	if (drawing) {
		let drawCoords = getDrawCoordinates(evt);

		draw(drawCoords.drawX, drawCoords.drawY);
		if (eraserActive)
			moveEraser(drawCoords.eraserX, drawCoords.eraserY);
	}
}

function mouseStop(evt) {
	if (!drawing)
		return;

	checkClickedOnTextarea(evt.clientX, evt.clientY);

	let drawCoords = getDrawCoordinates(evt);
	drawing = false;
	stopDraw(drawCoords.drawX, drawCoords.drawY);
	if (eraserActive)
		stopEraser();
}


// Touch Input

function touchStart(evt) {

	setContainerHeight(currentSlide);
	let touch = getPenInput(evt);
	if (touch === null)
		return;

	evt.preventDefault();

	drawID = touch.identifier;
	pos = computeCumulativeOffset(container);

	let drawCoords = getDrawCoordinates(touch);

	startDraw(drawCoords.drawX, drawCoords.drawY);
	if (eraserActive)
		initializeEraser(drawCoords.eraserX, drawCoords.eraserY);
}

function touchMove(evt) {

	let touch = getPenInput(evt);
	if (touch === null || touch.identifier !== drawID)
		return;

	evt.preventDefault();
	let drawCoords = getDrawCoordinates(touch);

	draw(drawCoords.drawX, drawCoords.drawY);
	if (eraserActive)
		moveEraser(drawCoords.eraserX, drawCoords.eraserY);
}

function touchStop(evt) {

	let touch = evt.changedTouches[0];

	if (touch.identifier !== drawID)
		return;

	evt.preventDefault();
	let drawCoords = getDrawCoordinates(touch);

	checkClickedOnTextarea(touch.clientX, touch.clientY);

	stopDraw(drawCoords.drawX, drawCoords.drawY);
	if (eraserActive)
		stopEraser();
}

function getPenInput(evt) {

	let touch = null;
	for (let i = 0; i < evt.targetTouches.length; i++) {
		if (evt.targetTouches[i].force > 0 || fingerDraw) {
			touch = evt.targetTouches[i];
			break;
		}
	}
	return touch;
}

function checkClickedOnTextarea(x, y) {
	clickedOnTextarea = false;
	for (let i = 0; i < textAreaArray[currentUser].length; i++) {

		if (checkIfContains(textAreaArray[currentUser][i].div, x, y)) {
			clickedOnTextarea = true;
			break;
		}
	}
}

//
// Zoom functions
//

function checkDoubleTab(x, y) {

	let now = new Date().getTime();
	let timesince = now - doubleTabStart;
	let distance = Math.sqrt((doubleTabCoords[0] - x) * (doubleTabCoords[0] - x)
		+ (doubleTabCoords[1] - y) * (doubleTabCoords[1] - y));

	if (timesince < 300 && timesince > 0 && distance < 50) {
		return true;
	} else {
		doubleTabStart = now;
		doubleTabCoords = [x, y];
		return false;
	}
}

function zoomSlide(evt) {
	evt.preventDefault();

	currentScale = currentScale * Math.max(0.95, Math.min(1.05, evt.scale));
	currentScale = Math.max(1, Math.min(5, currentScale));

	currentTranslate[0] = -zoomStart[0] * (1 - 1 / currentScale);
	currentTranslate[1] = -zoomStart[1] * (1 - 1 / currentScale);
	setScaleAndTranslate(allSlides[currentSlide], currentScale, currentTranslate[0], currentTranslate[1]);
}

function translateSlide(evt) {

	evt.preventDefault();
	evt = evt.targetTouches[0];

	let coords = getRelativeCoordinates(evt.pageX, evt.pageY);

	let left = coords.x / container.clientWidth;
	let top = coords.y / container.clientHeight;

	currentTranslate[0] = translateX + (left - translateStart[0]) * 100 / currentScale;
	currentTranslate[1] = translateY + (top - translateStart[1]) * 100 / currentScale;

	currentTranslate[0] = clamp(currentTranslate[0], -100 * (1 - 1 / currentScale), 0);
	currentTranslate[1] = clamp(currentTranslate[1], -100 * (1 - 1 / currentScale), 0);

	setScaleAndTranslate(allSlides[currentSlide], currentScale, currentTranslate[0], currentTranslate[1]);
}

function setScaleAndTranslate(element, scale, x, y) {
	element.style.transform = 'scale(' + scale + ') translate(' + x + '%, ' + y + '%)';
}

function clamp(num, min, max) {
	return num <= min ? min : num >= max ? max : num;
}

//
// Slider functions
//

function handleTouchStart(evt) {

	// Disable touch slider for finger draw mode
	if (fingerDraw)
		return;

	if (evt.targetTouches[0].force > 0)
		return;

	document.activeElement.blur();

	removeSlideTransitionEffect();

	pos = computeCumulativeOffset(container);

	if (evt.targetTouches.length > 1) {
		let coords = getRelativeCoordinates(evt.pageX, evt.pageY);

		zoomStart = [
			-translateX + coords.x / container.clientWidth * 100 / currentScale,
			-translateY + coords.y / container.clientHeight * 100 / currentScale
		];
		isZooming = true;
		evt.preventDefault();
	} else if (checkDoubleTab(evt.pageX, evt.pageY)) {
		isZooming = false;
		isTranslating = false;
		currentScale = 1;
		translateX = 0;
		translateY = 0;
		addSlideTransitionEffect();
		setScaleAndTranslate(allSlides[currentSlide], currentScale, translateX, translateY);
	}
	else if (currentScale !== 1) {
		let coords = getRelativeCoordinates(evt.pageX, evt.pageY);
		translateStart = [
			coords.x / container.clientWidth,
			coords.y / container.clientHeight
		];
		isTranslating = true;
	} else {
		slideStart(evt);
	}

}

function handleTouchMove(evt) {

	// Disable touch slider for finger draw mode
	if (fingerDraw)
		return;

	if (evt.targetTouches[0].force > 0)
		return;

	if (isZooming)
		zoomSlide(evt);
	else if (isTranslating)
		translateSlide(evt);
	else
		slideMove(evt);
}

function handleTouchEnd(evt) {

	// Disable touch slider for finger draw mode
	if (fingerDraw)
		return;

	addSlideTransitionEffect();

	if (isZooming) {
		translateX = currentTranslate[0];
		translateY = currentTranslate[1];
		isZooming = false;
	} else if (isTranslating) {
		translateX = currentTranslate[0];
		translateY = currentTranslate[1];
		isTranslating = false;
	} else {
		slideEnd(evt);
	}

	dragStart = -1;
	dragID = -1;
}

function addSlideTransitionEffect() {
	for (let i = 1; i <= maxSlide; i++) {
		allSlides[i].classList.add('slideEffect');
	}
}


function removeSlideTransitionEffect() {
	for (let i = 1; i <= maxSlide; i++) {
		allSlides[i].classList.remove('slideEffect');
	}
}

function slideStart(evt) {

	let touch = evt.targetTouches[0];
	if (touch.force > 0 || currentScale !== 1) {
		dragStart = -1;
		dragID = -1;
		return;
	}

	if (dragID > 0)
		return;

	dragID = touch.identifier;
	dragStart = touch.pageX;
	dragTime = new Date().getTime();
}

function slideMove(evt) {

	// Filter for event with correct touch id
	let touches = evt.targetTouches;
	let touch = null;

	for (let i = 0; i < touches.length; i++) {
		if (touches[i].identifier === dragID) {
			touch = touches[i];
		}
	}

	if (touch === null) {
		return;
	}

	if (dragStart < 0) {
		return;
	}

	let dragOffset = touch.pageX - dragStart;
	let ratio = dragOffset / container.clientWidth * 100; // * 1000 / 100

	if (dragOffset > 50 || dragOffset < -50) {
		evt.preventDefault();
	}

	if (ratio < 0) { // slide left

		if (currentSlide < maxSlide) {
			allSlides[currentSlide + 1].style.transform = 'translate(' + (ratio + 101) + '%,0)';
			allSlides[currentSlide].style.transform = 'translate(' + ratio + '%,0)';
		} else {
			allSlides[currentSlide].style.transform = 'translate(' + (ratio / 10) + '%,0)';
		}
	} else { // slide right
		if (currentSlide > 1) {
			allSlides[currentSlide - 1].style.transform = 'translate(' + (ratio - 101) + '%,0)';
			allSlides[currentSlide].style.transform = 'translate(' + ratio + '%,0)';
		} else {
			allSlides[currentSlide].style.transform = 'translate(' + (ratio / 10) + '%,0)';
		}

	}

}

function slideEnd(evt) {

	let touches = evt.changedTouches;
	let touch = null;

	for (let i = 0; i < touches.length; i++) {
		if (touches[i].identifier === dragID) {
			touch = touches[i];
		}
	}

	if (dragStart < 0 || touch === null)
		return;

	let dragOffset = touch.pageX - dragStart;
	let ratio = dragOffset / container.clientWidth * 100;
	let dragSpeed = ratio / (new Date().getTime() - dragTime);

	let pageChangeRatio = 20;
	let pageChangeSpeed = 0.1;

	if (ratio < -pageChangeRatio || dragSpeed < -pageChangeSpeed)
		gotoPage(currentSlide + 1);
	else if (ratio > pageChangeRatio || dragSpeed > pageChangeSpeed)
		gotoPage(currentSlide - 1);
	else {
		gotoPage(currentSlide);
	}
}

function handleChangePage() {
	setToolbarIcons();
	attachCanvas(currentSlide);
	setContainerHeight(currentSlide);
	setPagination();

	// Reset Zoom & Translate Values
	translateX = 0;
	translateY = 0;
	currentScale = 1;
	isZooming = false;
	isTranslating = false;
}

//
// Change Pencil Size
//

function setPenSizeCanvas(canvas, newSize, eraser) {
	if (!eraser) {
		canvas.getContext('2d').lineWidth = newSize * 2;
	} else {
		canvas.getContext('2d').lineWidth = newSize * 2 * eraserSizeFactor;
	}

}

function setPenSize(newSize) {
	currentPenSize = newSize;

	setPenSizeCanvas(mainCanvas[correctionPass], newSize, eraserActive);
}

//
// Switch Eraser and Pen
//

function setEraserContextCanvas(canvas, eraser, size) {
	if (eraser) { // Eraser activated
		canvas.getContext('2d').globalCompositeOperation = 'destination-out';
	} else {
		canvas.getContext('2d').globalCompositeOperation = 'source-over';
	}
	setPenSizeCanvas(canvas, size, eraser);
}

function setEraserContext(on) {

	eraserActive = on;

	setEraserContextCanvas(mainCanvas[correctionPass], on, currentPenSize);
}

//
// Save image
//

function saveData() {
	if (textMode)
		toggleTextMode();

	let output = {};
	output.images = [];

	let canvas = document.createElement('canvas');
	let emptyCanvas = document.createElement('canvas');

	for (let i = 1; i <= maxSlide; i++) {
		if (undoArray[correctionPass][i].length <= 0)
			continue;

		canvas.width = imgSize[i].width;
		canvas.height = imgSize[i].height;
		emptyCanvas.width = imgSize[i].width;
		emptyCanvas.height = imgSize[i].height;
		setupContext(canvas.getContext('2d'), correctionPass);
		replayUndoArraySlide(canvas, i, correctionPass);
		printInitialImage(emptyCanvas, i, correctionPass);
		let data = canvas.toDataURL('image/png');
		let emptyCanvasData = emptyCanvas.toDataURL('image/png');
		if (data === emptyCanvasData) {
			continue;
		}
		undoArray[correctionPass][i] = [];
		redoArray[correctionPass][i] = [];
		initialLayer[correctionPass][i].src = data;
		data = data.replace(/^data:image\/\w+;base64,/, '');
		output.images.push({'page': i, 'data': data});
	}

	return output;
}

//
// Undo
//

function Path(penSize, eraser) {
	this.penSize = penSize;
	this.eraser = eraser;
	this.coords = []; // [ [x1,y1], [x2, y2], ...]
}

function startNewPath() {
	stopPath();
	currentPath = null;
}

function stopPath() {
	if (currentPath !== null) {
		addUndoElement(new UndoPath(currentPath.eraser, currentPath.penSize, currentPath.coords));
	}
	currentPath = null;
}

function addUndoElement(element, page = currentSlide) {

	undoArray[currentUser][page].push(element);
	redoArray[currentUser][page].length = 0;   // Empty the Redo list
	setToolbarIcons();

	if (element.type !== 'path') {
		textAreaCounter[currentUser][element.id]++;
	}
}

function addPointToPath(path, x, y) {
	path.coords.push([x, y]);
}

function undo() {
	stopPath();
	if (undoArray[currentUser][currentSlide].length > 0) {
		let obj = undoArray[currentUser][currentSlide].pop();
		redoArray[currentUser][currentSlide].push(obj);

		//TODO
		if (obj.type === 'create') {
			textAreaArray[currentUser][obj.id].div.style.top = '5000%';
		}
		if (obj.type !== 'path') {
			textAreaCounter[currentUser][obj.id]--;
		}
	}
	mainContext[currentUser].clearRect(0, 0, imgSize[currentSlide].width, imgSize[currentSlide].height);
	replayUndoArraySlide(mainCanvas[currentUser], currentSlide, currentUser);

	setEraserContextCanvas(mainCanvas[currentUser], eraserActive, currentPenSize);
	setToolbarIcons();
}

function redo() {

	if (redoArray[currentUser][currentSlide].length === 0)
		return;

	let obj = redoArray[currentUser][currentSlide].pop();

	if (obj.type !== 'path') {
		textAreaCounter[currentUser][obj.id]++;
	}
	mainContext[currentUser].clearRect(0, 0, imgSize[currentSlide].width, imgSize[currentSlide].height);
	undoArray[currentUser][currentSlide].push(obj);
	replayUndoArraySlide(mainCanvas[currentUser], currentSlide, currentUser);

	setEraserContextCanvas(mainCanvas[currentUser], eraserActive, currentPenSize);
	setToolbarIcons();
}

function setTextAreaCounter(pass) {
	for (let i = 0; i < textAreaCounter[pass].length; i++) {
		textAreaCurrentCounter[i] = textAreaCounter[pass][i];
	}
}

function printInitialImage(canvas, pageNumber, correction_pass) {
	let eraserState = canvas.getContext('2d').globalCompositeOperation;
	canvas.getContext('2d').globalCompositeOperation = 'source-over';
	canvas.getContext('2d').drawImage(initialLayer[correction_pass][pageNumber], 0, 0);
	canvas.getContext('2d').globalCompositeOperation = eraserState;
}

function replayUndoArraySlide(canvas, pageNumber, pass) {
	printInitialImage(canvas, pageNumber, pass);
	setTextAreaCounter(pass);
	let eraser = eraserActive;
	setEraserContext(false);
	for (let i = 0; i < undoArray[pass][pageNumber].length; i++) {
		replayUndoObj(canvas, undoArray[pass][pageNumber][i], pageNumber, pass);
	}
	setEraserContext(eraser);
}

function replayUndoObj(canvas, obj, pageNumber, pass) {

	let type = obj.type;
	switch (type) {
		case 'path':
			drawPath(canvas, obj);
			break;
		case 'create':
			textAreaArray[pass][obj.id].div.style.left = obj.x + '%';
			textAreaArray[pass][obj.id].div.style.top = obj.y + '%';
			textAreaArray[pass][obj.id].textarea.value = obj.string;
			break;
		case 'delete':
			textAreaArray[pass][obj.id].textarea.value = '';
			textAreaArray[pass][obj.id].div.style.top = '200%';
			break;
		case 'move':
			textAreaArray[pass][obj.id].div.style.left = obj.newX + '%';
			textAreaArray[pass][obj.id].div.style.top = obj.newY + '%';
			break;
		case 'change':
			textAreaArray[pass][obj.id].textarea.value = obj.string;
			textAreaArray[pass][obj.id].resizeTextarea();
			break;
	}

	if (type !== 'path') {

		textAreaCurrentCounter[obj.id]--;

		if (textAreaCurrentCounter[obj.id] === 0 && !textMode) {
			printTextarea(textAreaArray[pass][obj.id], canvas, pageNumber);
		}
	}
}

String.prototype.replaceAll = function (find, replace) {
	let result = this;
	let split = [];
	do {
		split = result.split(find);
		result = split.join(replace);
	} while (split.length > 1);
	return result;
};

function drawPath(canvas, path) {

	setEraserContextCanvas(canvas, path.eraser, path.penSize);

	let coords = path.coords;
	let context = canvas.getContext('2d');

	let i;
	let cx = coords[0][0];
	let cy = coords[0][1];

	context.beginPath();
	context.moveTo(cx, cy);

	for (i = 1; i < coords.length - 1; i++) {
		cx = (coords[i][0] + coords[i + 1][0]) / 2;
		cy = (coords[i][1] + coords[i + 1][1]) / 2;
		context.quadraticCurveTo(coords[i][0], coords[i][1], cx, cy);
	}

	context.quadraticCurveTo(cx, cy, coords[i][0], coords[i][1]);
	context.stroke();
	context.beginPath();
}

//
// Change mode draw/write
//

function toggleTextMode() {

	// In case of a textarea is currently active
	document.activeElement.blur();

	textMode = !textMode;
	// Toggle display of all textareas
	for (let i = 0; i < textAreaArray[currentUser].length; i++) {
		if (i === currentCutTextarea)
			continue;
		$(textAreaArray[currentUser][i].div).toggleClass('d-none', !textMode) // hide when not in textmode
	}

	$('#textMode').prop('checked', textMode);

	// Redraw all Elements
	mainCanvas[currentUser].getContext('2d').clearRect(0, 0, imgSize[currentSlide].width, imgSize[currentSlide].height);
	replayUndoArraySlide(mainCanvas[currentUser], currentSlide, currentUser);
}

function printTextarea(textarea, canvas, pageNumber) {
	let context = canvas.getContext('2d');

	let imageScale = Math.max(imgSize[pageNumber].width, imgSize[pageNumber].height) / 3500;

	let offsetLeft = imageScale * 15;
	let offsetTop = imageScale * 150;
	let lineHeight = imageScale * defaultFontSize;

	let x = offsetLeft + (parseFloat(textarea.div.style.left) / 100) * imgSize[pageNumber].width;
	let y = offsetTop + (parseFloat(textarea.div.style.top) / 100) * imgSize[pageNumber].height;

	let subStrings = textarea.textarea.value.split('\n');

	// Deactivate eraser if necessary
	let eraser = (context.globalCompositeOperation === 'destination-out');
	if (eraser)
		context.globalCompositeOperation = 'source-over';

	// write textare content on context
	for (let i = 0; i < subStrings.length; i++) {
		context.fillText(subStrings[i], x, y);
		y += lineHeight;
	}
	context.stroke();

	// Activate erases if it was activated before
	if (eraser)
		context.globalCompositeOperation = 'destination-out';
}

//
// Generate click box svg
//

document.createSvg = function (tagName) {
	let svgNS = 'http://www.w3.org/2000/svg';
	return this.createElementNS(svgNS, tagName);
};

//
// GUI helper functions
//

function disableGuiOnDraw() {
	$('#toolbar, #pagination, .crossBox, .highlightProblem').css('opacity', .1).css('pointer-events', 'none')
}

function enableGuiOnDraw() {
	$('#toolbar, #pagination, .crossBox, .highlightProblem').css('opacity', 1).css('pointer-events', 'initial')
}

function setToolbarIcons() {
	$('#slideLeft').toggleClass('disabled', currentSlide === 1);
	$('#slideRight').toggleClass('disabled', currentSlide === maxSlide);

	$('#undo').toggleClass('disabled', undoArray[currentUser][currentSlide].length === 0)
	$('#redo').toggleClass('disabled', redoArray[currentUser][currentSlide].length === 0)
}

function readonlyToolbar(){
	$('#toolbar').find(
		'#undo, #redo, #createTextArea, #fingerDraw, #save'
	).addClass('disabled'); // for buttons
	$('#toolbar').find(
		'#selectPen, #selectEraser, #textMode, #fingerDraw'
	).prop('disabled', true); // For checkboxes and radio
}

function lockKeepAlive() {

	// TODO
	/*
	let url = Flask.url_for('exam_lock.api.lock',
		{
			exam_id: exam_data.exam,
			erid: exam_data.erid
		});

	let options = {
		method: 'PUT',
		mode: 'same-origin',
		credentials: 'include',
		headers: {
            'Content-Type': 'application/json',
			'X-CSRF-Token': $('meta[name=csrf-token]').attr('content'),
        },
		body: JSON.stringify({token: lockToken}),
	};

	// TODO: find better solution without alert
	fetch(url, options).then(function (res) {
		if (!res.ok)
			throw res;
	}).then(data => {
		lockRetries = 0;
		updateLog(iconLockClosed,`Exam lock updated.`, 'alert-success');
		setTimeout(lockKeepAlive, keepAliveInterval * 1000);
	}).catch(err => {
		if (err.status === 423) {
			updateLog(iconLockOpen, `Exam has been unlocked. Please reload.`, 'alert-danger');
			// TODO: Remove all eventListeners
		} else if (err.status === 401 || err.status === 403 || err.status === 503) {
			updateLog(iconLockOpen, `Your session has expired. Please reload the page.`, 'alert-danger');
		} else {
			lockRetries += 1;
			if (lockRetries < lockMaxRetries) {
				let wait = keepAliveInterval / lockMaxRetries;
				updateLog(iconLockOpen, `Could not contact Server. Trying again in ${wait} seconds. (${lockRetries}/${lockMaxRetries} retries)`, 'alert-warning');
				setTimeout(lockKeepAlive, wait * 1000);
			} else {
				updateLog(iconClose, `Connection to server lost. Unsaved changes are possibly lost.`, 'alert-danger');
			}
		}
	});
	 */
}

function updateLog(msgShort, msg, cls) {
	let ts = new Date($.now());
	ts = ts.getHours() + ":" + String(ts.getMinutes()).padStart(2, '0') + ":" + String(ts.getSeconds()).padStart(2, '0');
	$('#toolbarLogModal tbody').append(
		$(`<tr>
			<td>${ts}</td>
			<td>${msgShort}</td>
			<td>${msg}</td>
		</tr>`)
	)
	let elem = $('#toolbarLog');
	elem.find('span').html(msgShort);
	elem.removeClass ((index, className) => {
		return (className.match (/(^|\s)alert-\S+/g) || []).join(' ');
	});
	elem.addClass(cls);
}

//
// Pagination
//

function createPagination() {
	let div = document.getElementById('pageList');

	let elem = $(`<li class="page-item" id="slideLeft">
		<a class="page-link">
		 	<span aria-hidden="true">&laquo;</span>
		</a>
	</li>`)
	$(div).append(elem);

	elem.on('click', function () {
		addSlideTransitionEffect();
		gotoPage(currentSlide - 1);
	});


	for (let i = 1; i <= maxSlide; i++) {

		elem = $(`<li class="page-item" value="${i}">
			<a class="page-link">
		 		${i}
			</a>
		</li>`)
		$(div).append(elem);

		elem.on('click', function () {
			gotoPage(i);
		});
	}

	elem = $(`<li class="page-item" id="slideRight">
		 <a class="page-link">
		 	<span aria-hidden="true">&raquo;</span>
		</a>
	</li>`)
	$(div).append(elem);
	elem.on('click', function () {
		addSlideTransitionEffect();
		gotoPage(currentSlide + 1);
	});
}

function gotoPage(page) {
	// Clamp page to [1, maxslide]
	page = Math.min(maxSlide, Math.max(1, page));

	for (let i = 1; i <= maxSlide; i++) {
		if (i < page) {
			allSlides[i].style.transform = 'translate(-100%,0)';
		} else if (i > page) {
			allSlides[i].style.transform = 'translate(100%,0)';
		} else {
			allSlides[i].style.transform = 'translate(0,0)';
		}
	}
	currentSlide = page;
	loadSlide(page, false);
	loadSlide(page + 1, false);
	loadSlide(page - 1, false);
	handleChangePage();
}

function setPagination() {
	$.each($('#pageList .page-item'), (i, page) => {
		$(page).toggleClass('active', parseInt($(page).attr('value')) === currentSlide);
	});
}

function toggleFullscreen() {

	// This add additional offset to the toolbar in case to top navbar is displayed
	let navbar_height = parseInt(window.getComputedStyle(document.body).getPropertyValue('--navbar-top-height'));
	$('#toolbar').css('top', `${fullscreen ? 10 : 10 + navbar_height}px`)

	$('.top-nav, .exam-nav-placeholder, .exam-nav').toggleClass('d-none', fullscreen);
	gotoPage(currentSlide);
}

//
// Local Storage
//

function getCorrectionPreferences() {
	let storedData = JSON.parse(sessionStorage.getItem('correctionPreferences') || '{}') ;

	fullscreen = storedData.fullscreen || fullscreen;
	$('#fullscreenCheckbox').prop('checked', fullscreen);

	showIndividualSolution = storedData.showIndividualSolution || showIndividualSolution;
	$('#showIndividualSolution').prop('checked', showIndividualSolution);

	solutionOpacity = storedData.solutionOpacity || solutionOpacity;
	$('#solutionOpacity').val(solutionOpacity * 100);

	fingerDraw = storedData.fingerDraw || fingerDraw;
	$('#fingerDraw').prop('checked', fingerDraw);
}

function setCorrectionPreferences() {

	let data = {
		fullscreen: fullscreen,
		solutionOpacity: solutionOpacity,
		showIndividualSolution: showIndividualSolution,
		fingerDraw: fingerDraw,
	}
	sessionStorage.setItem('correctionPreferences', JSON.stringify(data));
}

//
// All added event listeners
//

function addEventListenersToCanvas() {

	mainCanvas[correctionPass].addEventListener('mousemove', mouseMove);
	mainCanvas[correctionPass].addEventListener('mousedown', mouseStart);
	mainCanvas[correctionPass].addEventListener('mouseup', mouseStop);
	mainCanvas[correctionPass].addEventListener('mouseleave', mouseStop);

	mainCanvas[correctionPass].addEventListener('touchstart', touchStart);
	mainCanvas[correctionPass].addEventListener('touchmove', touchMove);
	mainCanvas[correctionPass].addEventListener('touchend', touchStop);
}

function addSliderEventListener() {
	container.addEventListener('touchstart', handleTouchStart);
	container.addEventListener('touchmove', handleTouchMove);
	container.addEventListener('touchend', handleTouchEnd);
}

function removeSliderEventListener() {
	container.removeEventListener('touchstart', handleTouchStart);
	container.removeEventListener('touchmove', handleTouchMove);
	container.removeEventListener('touchend', handleTouchEnd);
}

function addEventListeners() {

	$('#correctionPreferences').on('click', evt => {
		$('#correctionPreferencesModal').modal('show');
	});

	addSliderEventListener();

	$('#fullscreenCheckbox').change(function() {
		fullscreen = $(this).is(':checked');
		setCorrectionPreferences();
		toggleFullscreen();
	});

	$('#showIndividualSolution').change(function() {
		showIndividualSolution = $(this).is(':checked');
		setCorrectionPreferences();
		toggleFullscreen();
	});

	$('#solutionOpacity').on('change', evt => {
		let opacity = $('#solutionOpacity').val() / 100;
		solutionOpacity = opacity < 1 ? opacity * 0.8 + 0.1 : opacity; // this makes the solution opacity in range [.1, .9] + [1]
		setCorrectionPreferences();
		$('.solutionImage').css('opacity', solutionOpacity).toggleClass('opaque', solutionOpacity < 1);
	});

	if(isReadonly)
		return;

	addEventListenersToCanvas();

	//Eraser
	let penButton = document.getElementById('selectPen');
	let eraserButton = document.getElementById('selectEraser');

	penButton.addEventListener('click', function () {
		setEraserContext(false);
	});

	eraserButton.addEventListener('click', function () {
		setEraserContext(true);
	});

	// Undo
	let undoButton = document.getElementById('undo');
	undoButton.addEventListener('click', function () {
		if (undoArray[currentUser][currentSlide].length === 0)
			return;

		undo();
	});

	// Redo
	let redoButton = document.getElementById('redo');
	redoButton.addEventListener('click', function () {
		if (redoArray[currentUser][currentSlide].length === 0)
			return;

		redo();
	});

	// Create Textarea
	let createTextArea = document.getElementById('createTextArea');
	createTextArea.addEventListener('click', function () {
		//TODO support multiple problems
		let prob = problems[currentProblem + currentSubproblem[0]];

		let left = 10;
		let top = 10;

		// shift textareas a little
		left += 0.1 * textAreaArray[currentUser].length;
		top += 0.1 * textAreaArray[currentUser].length;

		generateTextarea(currentProblem, currentSubproblem[0], left, top);
	});

	// Activate textarea mode
	$('#textMode').change(function() {
		toggleTextMode();
	});

	// Activate finger draw mode
	$('#fingerDraw').change(function() {
		fingerDraw = $(this).is(':checked');
		setCorrectionPreferences();
	});

	$('#save').on('click', evt => {
		$('#saveSuccessfulModal .save-running').toggleClass('d-none', true);
		$('#saveSuccessfulModal .save-successful').toggleClass('d-none', true);
		$('#saveSuccessfulModal .save-failed').toggleClass('d-none', true);
		$('#saveSuccessfulModal').modal('show');

		if (saveIsRunning)
			return

		saveIsRunning = true;
		$('#save').html(iconSpinnerSmall);
		$('#saveSuccessfulModal .save-running').toggleClass('d-none', false);

		let data = saveData();

		let url = '/save'

		let options = {
			method: 'POST',
			mode: 'same-origin',
			cache: 'no-cache',
			credentials: 'include',
			headers: {
			  	'Content-Type': 'application/json',
				'Accept': 'application/pdf'
			},
			body: JSON.stringify(data),
		};

		fetch(url, options)
			.then( res => {
				if (!res.ok)
					throw res;
				return res.blob()
		}).then(blob => {

			let file = window.URL.createObjectURL(blob);
			window.open(file, '_blank');

			$('#saveSuccessfulModal .save-running').toggleClass('d-none', true);
			$('#saveSuccessfulModal .save-successful').toggleClass('d-none', false);
			$('#save').html(iconSave);
			saveIsRunning = false;

		}).catch(res => {
			$('#saveSuccessfulModal .save-running').toggleClass('d-none', true);
			$('#saveSuccessfulModal .save-failed').toggleClass('d-none', false);
			$('#saveSuccessfulModal .save-failed .log').html(`${res.status} ${res.statusText}`)
			$('#save').html(iconSave);
			saveIsRunning = false;
			console.log(res);
		});
	});

	// Handle orientation change
	document.addEventListener('orientationchange', function(e) {
		setContainerHeight(currentSlide);
	});

	// Handle window resizing
	window.addEventListener('resize', function(e) {
		setContainerHeight(currentSlide);
	});

	document.addEventListener('touchstart', function (event) {
		if (event.scale !== 1) {
			event.preventDefault();
		}
	}, false);

}
